/*
 * Copyright (C) 2012 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by Pawel Stolowski <pawel.stolowski@canonical.com>
 *
 */
using Gtk;

namespace Unity.Tester {

public class UnityToolUi: GLib.Object
{
    public UnityToolUi()
    {
    }

    public bool init_gui()
    {
        var builder = new Builder ();

        try {
            builder.add_from_resource("/com/canonical/Unity/unity-tool/unity-tool.ui");
            builder.connect_signals(this);

            spinner = builder.get_object ("spinner") as Spinner;

            var window = builder.get_object ("window") as Window;
            window.destroy.connect(Gtk.main_quit);
            window.show_all();

            uimodel = builder.get_object("results_model") as ListStore;
            ui_filter_model = builder.get_object("filters_model") as ListStore;

            search_entry = builder.get_object("search_entry") as Entry;
            search_type_global_rbutton = builder.get_object("search_type_global") as RadioButton;
            search_button = builder.get_object("search_button") as Button;
            statusbar = builder.get_object("statusbar") as Statusbar;
            log_buffer = builder.get_object("log_buffer") as TextBuffer;

            statusbar_info_ctx = statusbar.get_context_id("Info");
            statusbar_error_ctx = statusbar.get_context_id("Error");

            show_connect_dialog();
        }
        catch (GLib.Error e) {
            ui_load_error(e.message);
			return false;
        }
        return true;
    }

	private void ui_load_error(string message)
	{
		Gtk.Dialog dlg = new Gtk.MessageDialog(null,
                    Gtk.DialogFlags.DESTROY_WITH_PARENT | Gtk.DialogFlags.MODAL,
                    Gtk.MessageType.ERROR, Gtk.ButtonsType.CLOSE,
                    "Error loading UI file:\n%s".printf(message));
		dlg.title = "Error creating UI";
		dlg.run();
		dlg.destroy();
	}

    private void discover_lens()
    {
        lens_discovery_spinner.start();

        TreeIter iter;
        lens_list_model.append(out iter);
        lens_list_model.set(iter, 0,"", 1, "", -1); //add empty combobox entry

        DBusLensUtil c = new DBusLensUtil();
        c.findLenses.begin((obj, res) => {
                try {
                    unowned List<DBusLensUtil.DBusObjectAddress?> results = c.findLenses.end(res);
                    foreach (DBusLensUtil.DBusObjectAddress addr in results) {
                        lens_list_model.append(out iter);
                        lens_list_model.set(iter, 0, addr.dbus_name, 1, addr.dbus_path, -1);
                    }
                    lens_discovery_spinner.stop();
                }
                catch (Error e) {
                    stderr.printf("DBus Lens auto-discovery failed %s\n", e.message);
                    lens_discovery_spinner.stop();
                }
        });
    }

    [CCode (instance_pos = -1)]
    public void on_lens_combobox_clicked(Gtk.MenuItem item)
    {
        TreeIter iter;
        if (lens_list_combobox.get_active() > 0) { // ignore first empty item
            if (lens_list_combobox.get_active_iter(out iter)) {
                Value val;
                lens_list_model.get_value(iter, 0, out val);
                dbus_name_entry.text = val.get_string();
                lens_list_model.get_value(iter, 1, out val);
                dbus_path_entry.text = val.get_string();
            }
        }
    }

    //
    // 'New' menu action handler
    [CCode (instance_pos = -1)]
    public void on_connect_clicked(Gtk.MenuItem item)
    {
        show_connect_dialog();
    }

    private void show_connect_dialog()
    {
        var builder = new Builder ();
		try {
			builder.add_from_resource("/com/canonical/Unity/unity-tool/dbus-lens-connect.ui");
			builder.connect_signals(this);
			lens_discovery_spinner = builder.get_object("lens_discovery_spinner") as Spinner;
			lens_list_model = builder.get_object("lens_list_model") as ListStore;
			lens_list_combobox = builder.get_object("lens_list_combobox") as ComboBox;
			dbus_name_entry = builder.get_object("dbus_name_entry") as Entry;
			dbus_path_entry = builder.get_object("dbus_path_entry") as Entry;
			lens_connect_dlg = builder.get_object("lens_connect_dialog") as Dialog;

			discover_lens();
			
			if (Options.lens_dbus_path != null && Options.lens_dbus_path != "" && Options.lens_dbus_name != null && Options.lens_dbus_name != "") {
				dbus_name_entry.text = Options.lens_dbus_name;
				dbus_path_entry.text = Options.lens_dbus_path;
			}
			lens_connect_dlg.show_all();
		} 
		catch (GLib.Error e) {
			ui_load_error(e.message);
		}
    }

    private void results_row_added_cb(Dee.Model model, Dee.ModelIter iter)
    {
        var row = model.get_row(iter);
        TreeIter uiiter;
        uimodel.append(out uiiter);

        uimodel.set(uiiter, 0, row[0].get_string(), 1, row[1].get_string(), 2, row[2].get_uint32(), 3, row[3].get_string(), 4, row[4].get_string(), 5,
                row[5].get_string(), 6, row[6].get_string(), -1);
    }

    private void filters_row_added_cb(Dee.Model model, Dee.ModelIter iter)
    {
        var row = model.get_row(iter);
        TreeIter uiiter;
        ui_filter_model.append(out uiiter);

        ui_filter_model.set(uiiter, 0, row[0].get_string(), 1, row[1].get_string (), 2, row[2].get_string (), 3, row[3].get_string (), 4, row[4].print(true),
             5, row[5].get_boolean (), 6, row[6].get_boolean (), 7, row[7].get_boolean(), -1);
    }

    private void update_status()
    {
        statusbar.pop(statusbar_info_ctx);
        statusbar.push(statusbar_info_ctx, "%u records".printf(dee_results_model.get_n_rows()) + ", DBus name: " + Options.lens_dbus_name + " path: " +
                Options.lens_dbus_path);
    }

    private void model_synchronized_cb()
    {
        spinner.stop();
        update_status();
        append_log_message("Search returned %u records\n".printf(dee_results_model.get_n_rows()));
        dee_results_model = null;
    }

    private void filter_model_synchronized_cb()
    {
        dee_filters_model = null;
    }

    /**
     * Triggered clicking 'Ok' in the connection dialog.
     */
    [CCode (instance_pos = -1)]
    public void on_lens_connect(Gtk.Dialog dlg, int response)
    {
        //
        // user clicked 'Ok' in the Lens Connect dialog
        if (response == 1) {
            Options.lens_dbus_name = dbus_name_entry.text;
            Options.lens_dbus_path = dbus_path_entry.text;
        }
        if (Options.lens_dbus_path == null || Options.lens_dbus_path.length == 0 || Options.lens_dbus_name == null || Options.lens_dbus_name.length == 0) {
            search_entry.sensitive = false;
            search_button.sensitive = false;
        } else {
            search_entry.sensitive = true;
            search_button.sensitive = true;
        }
        dlg.destroy();
    }

    /**
     * Triggered by clicking 'From lens file' button in the connection dialog.
     * Opens file browser letting the user pick a .lens file.
     */
    [CCode (instance_pos = -1)]
    public void on_from_lens_clicked(Gtk.Button btn)
    {
        var filter = new FileFilter();
        filter.set_name("Lens files");
        filter.add_pattern("*.lens");

        var file_chooser = new FileChooserDialog("Open Lens file", null, Gtk.FileChooserAction.OPEN, Gtk.Stock.CANCEL, 0, Gtk.Stock.OPEN, 1);
        file_chooser.set_filter(filter);
        if (file_chooser.run() == 1) {
			try {
				get_lens_params_from_file(file_chooser.get_filename());
				dbus_name_entry.text = Options.lens_dbus_name;
				dbus_path_entry.text = Options.lens_dbus_path;
			}
			catch (Error e) {
				Gtk.Dialog dlg = new Gtk.MessageDialog(null,
								   Gtk.DialogFlags.DESTROY_WITH_PARENT | Gtk.DialogFlags.MODAL,
								   Gtk.MessageType.ERROR, Gtk.ButtonsType.CLOSE,
								   "Error loading lens parameters from file:\n%s".printf(e.message));
				dlg.title = "Error loading lens file";
				dlg.run();
				dlg.destroy();
			}
        }
        file_chooser.destroy();
    }

    /**
     * Appends log message to 'Log' tab.
     */
    private void append_log_message(string message)
    {
        TextIter iter;
        log_buffer.get_end_iter(out iter);
        log_buffer.insert(ref iter, message, -1);
    }

    /**
     * Request lens info via debus, creates dee models.
     */
    private void request_lens_info()
    {
        spinner.start();

        if (dee_filters_model != null) {
            dee_filters_model = null;
        }
        if (dee_results_model != null) {
            dee_results_model = null;
        }
        uimodel.clear();
        ui_filter_model.clear();

        LensInfo info = get_lens_info((msg) => { handle_error(msg); });
        string model_name = search_type_global_rbutton.get_active() ? info.global_results_model_name : info.results_model_name;
        dee_results_model = new Dee.SharedModel(model_name);
        model_sync_sig_id = dee_results_model.notify["synchronized"].connect(model_synchronized_cb);
        dee_results_model.row_added.connect(results_row_added_cb);

        dee_filters_model = new Dee.SharedModel(info.filters_model_name);
        filters_model_sync_sig_id = dee_filters_model.notify["synchronized"].connect(filter_model_synchronized_cb);
        dee_filters_model.row_added.connect(filters_row_added_cb);
    }

    [CCode (instance_pos = -1)]
    public void on_results_button_clicked(Gtk.Button btn)
    {
        request_lens_info();
    }

    /**
     * Triggered by clicking 'Search' button. Starts lens search over dbus.
     */
    [CCode (instance_pos = -1)]
    public void on_search_button_clicked(Gtk.Button btn)
    {
        string text = search_entry.text;
        if (text != "") {
            spinner.start();
       
            if (DBus.is_name(Options.lens_dbus_name) && GLib.Variant.is_object_path(Options.lens_dbus_path)) {
                append_log_message("Query: '%s' (%s), DBus name: %s, DBus path: %s\n".printf(
                            text,
                            search_type_global_rbutton.get_active() ? "global" : "local",
                            Options.lens_dbus_name,
                            Options.lens_dbus_path));
                call_lens_search(text, search_type_global_rbutton.get_active() ? 1 : 0, (result) =>
                        {
                            append_log_message(result.print(true) + "\n");
                            request_lens_info();
                        }, (msg) => {
                            handle_error(msg);
                        }
                    );
            } else {
                handle_error("Invalid DBus name/path");
            }
        }
    }

    /**
     * Helper method that stops spinner and puts error message on statusbar.
     */
    private void handle_error(string message)
    {
        spinner.stop();
        statusbar.pop(statusbar_error_ctx);
        statusbar.push(statusbar_error_ctx, message);
        append_log_message(message + "\n");
    }

    private Gtk.RadioButton search_type_global_rbutton = null;
    private Gtk.Dialog lens_connect_dlg = null;
    private Gtk.Spinner spinner = null;
    private Gtk.Spinner lens_discovery_spinner = null;
    private Gtk.Button search_button = null;
    private Gtk.Statusbar statusbar = null;
    private Gtk.TextBuffer log_buffer = null;
    private uint statusbar_info_ctx;
    private uint statusbar_error_ctx;
    private Gtk.Entry search_entry = null;
    private Gtk.Entry dbus_name_entry = null;
    private Gtk.Entry dbus_path_entry = null;
    private Gtk.ComboBox lens_list_combobox = null;
    private ListStore uimodel = null;
    private ListStore ui_filter_model = null;
    private ListStore lens_list_model = null;
    private ulong model_sync_sig_id;
    private ulong filters_model_sync_sig_id;
    private Dee.SharedModel? dee_results_model = null;
    private Dee.SharedModel? dee_filters_model = null;
}

}
